/*
 *
 *   OpenChanfix v2.0 
 *
 *   Channel Re-op Service Module for Hybrid 7.1 or later.
 *   Copyright (C) 2003 Thomas Carlsson and Joost Vunderink.
 *   See http://www.garion.org/ocf/ for more information.
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */



#include "cf_base.h"

static BlockHeap *ocf_user_heap  = NULL;
static BlockHeap *ocf_loggedin_user_heap  = NULL;
static char saltchars[] = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ./\0";

/* ----------------------------------------------------------------- */

void ocf_users_init()
{
	ocf_user_heap = BlockHeapCreate(sizeof(struct ocf_user), OCF_USER_HEAP_SIZE);
	ocf_loggedin_user_heap = BlockHeapCreate(sizeof(struct ocf_loggedin_user), OCF_LOGGEDIN_USER_HEAP_SIZE);
	
}

/* ----------------------------------------------------------------- */

void ocf_users_deinit()
{
	dlink_node *ptr, *next_ptr;
	DLINK_FOREACH_SAFE(ptr, next_ptr, ocf_user_list.head)
	{
		BlockHeapFree(ocf_user_heap, ptr->data);
		dlinkDelete(ptr, &ocf_user_list);
	}
	DLINK_FOREACH_SAFE(ptr, next_ptr, ocf_loggedin_user_list.head)
	{
		BlockHeapFree(ocf_loggedin_user_heap, ptr->data);
		dlinkDelete(ptr, &ocf_loggedin_user_list);
	}
}

/* ----------------------------------------------------------------- */

int ocf_users_get_unique_id()
{
	dlink_node *ptr;
	struct ocf_user *ou;
	int max = 0;

	DLINK_FOREACH(ptr, ocf_user_list.head) {
		ou = ((struct ocf_user*)(ptr->data));
		if (ou->id > max) {
			max = ou->id;
		}
	}

	return (max + 1);
}

/* ----------------------------------------------------------------- */

void ocf_users_updated()
{
	ocf_user_save_all(settings_get_str("userfile"));
}

/* ----------------------------------------------------------------- */

struct ocf_user* ocf_add_user(const char* name, const char* hostmask)
{
	struct ocf_user *user_p;

	if (ocf_get_user(name)) {
		/* User already exists */
		return NULL;
	}


	user_p = BlockHeapAlloc(ocf_user_heap);
	memset(user_p, 0, sizeof(struct ocf_user));
	strncpy(user_p->name, name, sizeof(user_p->name));
	strncpy(user_p->hostmasks, hostmask, sizeof(user_p->hostmasks));
	user_p->id = ocf_users_get_unique_id();
	dlinkAdd(user_p, make_dlink_node(), &ocf_user_list);

	return user_p;
}

/* ----------------------------------------------------------------- */

struct ocf_user* ocf_add_user_from_obj(struct ocf_user* other)
{
	struct ocf_user *user_p;

	if (ocf_get_user(other->name)) {
		/* User already exists */
		return NULL;
	}

	
	user_p = BlockHeapAlloc(ocf_user_heap);
	memcpy(user_p, other, sizeof(struct ocf_user));
	if (other->id < 1) {
		user_p->id = ocf_users_get_unique_id();
	}
	dlinkAdd(user_p, make_dlink_node(), &ocf_user_list);

	return user_p;
}

/* ----------------------------------------------------------------- */

void ocf_user_set_password(struct ocf_user* ocf_user_p, const char* password)
{
	char salt[3];
	salt[0] = saltchars[ (int) (random() % strlen(saltchars)) ];
	salt[1] = saltchars[ (int) (random() % strlen(saltchars)) ];
	salt[2] = 0;

	strncpy(ocf_user_p->password, crypt(password, salt), sizeof(ocf_user_p->password));
}

/* ----------------------------------------------------------------- */

int ocf_user_check_password(struct ocf_user* ocf_user_p, const char* password)
{
	char salt[3];
	char* passwordhash;

	salt[0] = ocf_user_p->password[0];
	salt[1] = ocf_user_p->password[1];
	salt[2] = 0;

	passwordhash = crypt(password, salt);
	if (!strncmp(passwordhash, ocf_user_p->password, sizeof(ocf_user_p->password)))
	{
		return 1;
	}

	return 0;
}

/* ----------------------------------------------------------------- */

int ocf_user_add_hostmask(struct ocf_user* ocf_user_p, const char* hostmask)
{
	/* Don't add the hostmask if that will exceed the maximum length for
	 * hostmasks. */
	if (strlen(ocf_user_p->hostmasks) + strlen(hostmask) + 1 > OCF_USER_HOSTS_LENGTH) {
		return 0;
	}

	/* Don't add the hostmask if it's already present */
	if (ocf_user_has_hostmask(ocf_user_p, hostmask)) {
		return 0;
	}

	if (strlen(ocf_user_p->hostmasks) == 0) {
		strncpy(ocf_user_p->hostmasks, hostmask, sizeof(ocf_user_p->hostmasks));
		return 1;
	}

	strcat(ocf_user_p->hostmasks, ",");
	strcat(ocf_user_p->hostmasks, hostmask);
	return 1;
}

/* ----------------------------------------------------------------- */

int ocf_user_del_hostmask(struct ocf_user* ocf_user_p, const char* hostmask)
{
	char *parv[MAXPARA];
	int i, parc;
	char masks[OCF_USER_HOSTS_LENGTH];

	if (!ocf_user_has_hostmask(ocf_user_p, hostmask)){
		return 0;
	}

	/* tokenize destroys the original string, so tokenize a copy */
	strncpy(masks, ocf_user_p->hostmasks, sizeof(masks));
	parc = tokenize_string_separator(masks, parv, ',');

	/* Re-add all hostmasks except the one that needs to be removed */
	ocf_user_p->hostmasks[0] = 0;
	for (i = 0; i < parc; i++) {
		if (strlen(hostmask) == strlen(parv[i])){
			if ( !strncmp(hostmask, parv[i], strlen(hostmask)) ) {
				/* don't re-add this host */
			} else {
				ocf_user_add_hostmask(ocf_user_p, parv[i]);
			}
		} else {
			ocf_user_add_hostmask(ocf_user_p, parv[i]);
		}
	}

	return 1;
}

/* ----------------------------------------------------------------- */

int ocf_user_has_hostmask(struct ocf_user* ocf_user_p, const char* hostmask)
{
	char *parv[MAXPARA];
	int i, parc;
	char masks[OCF_USER_HOSTS_LENGTH];
	
	if (strlen(ocf_user_p->hostmasks) == 0) {
		return 0;
	}

	/* tokenize destroys the original string, so tokenize a copy */
	strncpy(masks, ocf_user_p->hostmasks, sizeof(masks));
	parc = tokenize_string_separator(masks, parv, ',');

	for (i = 0; i < parc; i++) {
		/* TODO: wildcard host matching? */
		if (strlen(hostmask) == strlen(parv[i])){
			if ( !strncmp(hostmask, parv[i], strlen(hostmask)) ) {
				return 1;
			}
		}
	}

	return 0;
}

/* ----------------------------------------------------------------- */

void ocf_user_add_flags(struct ocf_user* ocf_user_p, int flags)
{
	ocf_user_p->flags |= flags;
}

/* ----------------------------------------------------------------- */

void ocf_user_del_flags(struct ocf_user* ocf_user_p, int flags)
{
	ocf_user_p->flags &= ~flags;
}

/* ----------------------------------------------------------------- */

int ocf_user_has_flag(struct ocf_user* ocf_user_p, int flag)
{
	/* Owner flag overrules all other flags. */
	if (ocf_user_p->flags & OCF_USER_FLAGS_OWNER) {
		return 1;
	}

	if (ocf_user_p->flags & flag) {
		return 1;
	}

	return 0;
}

/* ----------------------------------------------------------------- */

void ocf_user_add_notice_flags(struct ocf_user* ocf_user_p, int flags)
{
	ocf_user_p->noticeflags |= flags;
}

/* ----------------------------------------------------------------- */

void ocf_user_del_notice_flags(struct ocf_user* ocf_user_p, int flags)
{
	ocf_user_p->noticeflags &= ~flags;
}

/* ----------------------------------------------------------------- */

int ocf_user_has_notice_flag(struct ocf_user* ocf_user_p, int flag)
{
	if (ocf_user_p->noticeflags & flag) {
		return 1;
	}

	return 0;
}

/* ----------------------------------------------------------------- */

void ocf_user_disable(struct ocf_user* ocf_user_p)
{
	ocf_user_p->disabled = 1;
	/* Find out if the user is logged in; if so, log him out */
}

/* ----------------------------------------------------------------- */

void ocf_user_enable(struct ocf_user* ocf_user_p)
{
	ocf_user_p->disabled = 0;
}

/* ----------------------------------------------------------------- */

void ocf_user_delete(struct ocf_user* ocf_user_p)
{
	ocf_user_p->deleted = 1;
	/* Find out if the user is logged in; if so, log him out */
}

/* ----------------------------------------------------------------- */

char* ocf_user_get_name_from_id(int id)
{
	dlink_node *ptr;
	struct ocf_user *ou;

	DLINK_FOREACH(ptr, ocf_user_list.head) {
		ou = ((struct ocf_user*)(ptr->data));
		if (ou->id == id) {
			return ou->name;
		}
	}

	return NULL;
}

/* ----------------------------------------------------------------- */

void ocf_user_load_all(const char* filename)
{
	char line[MAX_LINE_LENGTH + 1]; /* Yeah, lazy. Should work with realloc perhaps. */
	FILE* fp;
	int linenum = 0;
	int numusers = 0;
	char *parv[MAXPARA];
	int parc;
	int i;
	struct ocf_user tmp_user;
	struct ocf_user *user_p;
	
    fp = fopen(filename, "r");

    if(!fp) {
		ocf_log(OCF_LOG_MAIN, "USER Error opening users file %s for reading.", filename);
		ocf_log(OCF_LOG_ERROR, "USER Error opening users file %s for reading.", filename);
        return;
    }

	memset(&tmp_user, 0, sizeof(tmp_user));

	/* Read lines until end of file.
	 * If a line contains 'id = <num>', add the previously gathered
	 * data as a new user. 
	 */
	while (readline(fp, line, MAX_LINE_LENGTH))
	{
        linenum++;
		strtrim(line);
        if (!*line || *line == '#')
            continue;

		parc = tokenize_string_separator(line, parv, '=');
		if (parc != 2) {
			/* log parse error on this line */
			continue;
		}

		strtrim(parv[0]);
		strtrim(parv[1]);
		if (strcmp(parv[0], "id") == 0) {
			if (numusers > 0) {
				user_p = ocf_add_user_from_obj(&tmp_user);
				if (!user_p) {
					ocf_log(OCF_LOG_MAIN, "Error adding user %s while loading users.", tmp_user.name);
					ocf_log(OCF_LOG_ERROR, "Error adding user %s while loading users.", tmp_user.name);
				}
				/* add previous, check for double name/id */
			}
			numusers++;
			memset(&tmp_user, 0, sizeof(tmp_user));
			tmp_user.id = atoi(parv[1]);
			/* add previous, start new user */
		}

		else if (strcmp(parv[0], "name") == 0) {
			strcpy(tmp_user.name, parv[1]);
		}

		else if (strcmp(parv[0], "pass") == 0) {
			strcpy(tmp_user.password, parv[1]);
		}

		else if (strcmp(parv[0], "flags") == 0) {
			for(i = 0; i < strlen(parv[1]); i++) {
				/* Ignore any minus characters */
				if (parv[1][i] != '-') {
					ocf_user_add_flags(&tmp_user, 
						ocf_user_flag_get_int(parv[1][i]));
				}
			}
		}

		else if (strcmp(parv[0], "umode") == 0) {
			for(i = 0; i < strlen(parv[1]); i++) {
				/* Ignore any minus characters */
				if (parv[1][i] != '-') {
					ocf_user_add_notice_flags(&tmp_user, 
						ocf_user_notice_flag_get_int(parv[1][i]));
				}
			}
		}

		else if (strcmp(parv[0], "host") == 0) {
			ocf_user_add_hostmask(&tmp_user, parv[1]);
		}

		else if (strcmp(parv[0], "deleted") == 0) {
			tmp_user.deleted = atoi(parv[1]);
		}

		else if (strcmp(parv[0], "disabled") == 0) {
			tmp_user.disabled = atoi(parv[1]);
		}
	}

	/* Add the final user */
	user_p = ocf_add_user_from_obj(&tmp_user);
	if (!user_p) {
		ocf_log(OCF_LOG_MAIN, "USER Error adding user %s while loading users.", tmp_user.name);
		ocf_log(OCF_LOG_ERROR, "USER Error adding user %s while loading users.", tmp_user.name);
	}

	fclose(fp);
}

/* ----------------------------------------------------------------- */

void ocf_user_save_all(const char* filename)
{
	FILE* fp;
	int numusers = 0;
	char hostmasks[OCF_USER_HOSTS_LENGTH + 1];
	char *parv[MAXPARA];
	int parc;
	int i;
	struct ocf_user *ou;
	dlink_node *ptr;

    fp = fopen(filename, "w");

    if(!fp) {
		ocf_log(OCF_LOG_MAIN, "USER Error opening users file %s for writing.", filename);
		ocf_log(OCF_LOG_ERROR, "USER Error opening users file %s for writing.", filename);
        return;
    }

	DLINK_FOREACH(ptr, ocf_user_list.head)
	{
		numusers++;
		ou = ((struct ocf_user*)(ptr->data));

		strncpy(hostmasks, ou->hostmasks, sizeof(hostmasks));
		parc = tokenize_string_separator(hostmasks, parv, ',');

		/* write user */
		writeline(fp, "id = %d", ou->id);
		writeline(fp, "name = %s", ou->name);
		writeline(fp, "pass = %s", ou->password);
		if (strlen(ocf_user_flags_to_string(ou->flags)) > 0) {
			writeline(fp, "flags = %s", ocf_user_flags_to_string(ou->flags));
		} else {
			writeline(fp, "flags = -");
		}
		if (strlen(ocf_user_notice_flags_to_string(ou->noticeflags)) > 0) {
			writeline(fp, "umode = %s", ocf_user_notice_flags_to_string(ou->noticeflags));
		} else {
			writeline(fp, "umode = -");
		}
		for (i = 0; i < parc; i++) {
			writeline(fp, "host = %s", parv[i]);
		}
		writeline(fp, "disabled = %d", ou->disabled);
		writeline(fp, "deleted = %d", ou->deleted);
		writeline(fp, "");
	}

	fclose(fp);
}

/* ----------------------------------------------------------------- */

struct ocf_loggedin_user* ocf_get_loggedin_user(const char* nick)
{
	dlink_node *ptr;
	struct ocf_loggedin_user *olu;

	DLINK_FOREACH(ptr, ocf_loggedin_user_list.head)
	{
		olu = ((struct ocf_loggedin_user*)(ptr->data));
		if (strcmp(olu->nick, nick) == 0)
		{
			return olu;
		}
	}
	
	return NULL;
}

/* ----------------------------------------------------------------- */

struct ocf_user* ocf_get_user(const char* name)
{
	dlink_node *ptr;
	struct ocf_user *ou;

	DLINK_FOREACH(ptr, ocf_user_list.head) {
		ou = ((struct ocf_user*)(ptr->data));
		if (strcasecmp(ou->name, name) == 0 && !ou->deleted)
			return ou;
	}
	
	return NULL;
}

/* ----------------------------------------------------------------- */

struct ocf_loggedin_user* ocf_user_login(const char* nick, const char *user, 
										 const char *host, const char *server,
										 struct ocf_user *u)
{
	struct ocf_loggedin_user *user_p;

	if (ocf_get_loggedin_user(nick)) {
		ocf_log(OCF_LOG_DEBUG, "oul %s already logged in", nick);
		/* User already exists */
		return NULL;
	}

	user_p = BlockHeapAlloc(ocf_loggedin_user_heap);
	memset(user_p, 0, sizeof(struct ocf_loggedin_user));
	strncpy(user_p->nick, nick, sizeof(user_p->nick));
	strncpy(user_p->user, user, sizeof(user_p->user));
	strncpy(user_p->host, host, sizeof(user_p->host));
	strncpy(user_p->server, server, sizeof(user_p->server));
	user_p->ocf_user_p = u;
	dlinkAdd(user_p, make_dlink_node(), &ocf_loggedin_user_list);

	return user_p;
}

/* ----------------------------------------------------------------- */

void ocf_user_logout(const char* nick)
{
	dlink_node *ptr;
	struct ocf_loggedin_user *olu;

	DLINK_FOREACH(ptr, ocf_loggedin_user_list.head)
	{
		olu = ((struct ocf_loggedin_user*)(ptr->data));
		if (strcmp( olu->nick, nick) == 0)
		{
			BlockHeapFree(ocf_loggedin_user_heap, ptr->data);
			dlinkDelete(ptr, &ocf_loggedin_user_list);
			return;
		}
	}
}

/* ----------------------------------------------------------------- */

void ocf_check_loggedin_users()
{
}

/* ----------------------------------------------------------------- */

const char* ocf_user_flags_to_string(int flags)
{
	static char flagchars[32];
	int pos = 0;
	int i;

	for (i = 1; i <= OCF_USER_FLAGS_LAST; i *= 2)
	{
		if (flags & i) {
			flagchars[pos] = ocf_user_flag_get_char(i);
			pos++;
		}
	}
	flagchars[pos] = 0;

	return flagchars;
}

/* ----------------------------------------------------------------- */

char ocf_user_flag_get_char(int flag)
{
	switch (flag) {
	case OCF_USER_FLAGS_OWNER      : return 'o';
	case OCF_USER_FLAGS_BLOCK      : return 'b';
	case OCF_USER_FLAGS_USERMANAGER: return 'u';
	case OCF_USER_FLAGS_FIX        : return 'f';
	}
	return ' ';
}

/* ----------------------------------------------------------------- */

char ocf_user_flag_get_int(char flag)
{
	switch (flag) {
	case 'o': return OCF_USER_FLAGS_OWNER;
	case 'b': return OCF_USER_FLAGS_BLOCK;
	case 'u': return OCF_USER_FLAGS_USERMANAGER;
	case 'f': return OCF_USER_FLAGS_FIX;
	}
	return 0;
}

/* ----------------------------------------------------------------- */

const char* ocf_user_notice_flags_to_string(int flags)
{
	static char flagchars[32];
	int pos = 0;
	int i;

	for (i = 1; i <= OCF_NOTICE_FLAGS_LAST; i *= 2)
	{
		if (flags & i) {
			flagchars[pos] = ocf_user_notice_flag_get_char(i);
			pos++;
		}
	}
	flagchars[pos] = 0;

	return flagchars;
}

/* ----------------------------------------------------------------- */

char ocf_user_notice_flag_get_char(int flag)
{
	switch (flag) {
	case OCF_NOTICE_FLAGS_BLOCK   : return 'b';
	case OCF_NOTICE_FLAGS_CHANFIX : return 'c';
	case OCF_NOTICE_FLAGS_LOGINOUT: return 'l';
	case OCF_NOTICE_FLAGS_NOTES   : return 'n';
	case OCF_NOTICE_FLAGS_USER    : return 'u';
	case OCF_NOTICE_FLAGS_AUTOFIX : return 'a';
	case OCF_NOTICE_FLAGS_MAIN    : return 'm';
	}
	return ' ';
}

/* ----------------------------------------------------------------- */

char ocf_user_notice_flag_get_int(char flag)
{
	switch (flag) {
	case 'b': return OCF_NOTICE_FLAGS_BLOCK;
	case 'c': return OCF_NOTICE_FLAGS_CHANFIX;
	case 'l': return OCF_NOTICE_FLAGS_LOGINOUT;
	case 'n': return OCF_NOTICE_FLAGS_NOTES;
	case 'u': return OCF_NOTICE_FLAGS_USER;
	case 'a': return OCF_NOTICE_FLAGS_AUTOFIX;
	case 'm': return OCF_NOTICE_FLAGS_MAIN;
	}
	return 0;
}

/* ----------------------------------------------------------------- */

